@@ -34,17 +34,21 @@
 #include <gnutls/x509.h>
 
 #include "gtlscertificate-gnutls.h"
+#include "gtlshttp.h"
+#include "gtlsgnutls-version.h"
 
 typedef struct
 {
   /*
    * This class is protected by mutex because the default GTlsDatabase
    * is a global singleton, accessible via the default GTlsBackend.
+   * Everything is read-only after initialization, but still has to
+   * be protected by the mutex.
    */
   GMutex mutex;
 
-  /* Read-only after construct, but still has to be protected by the mutex. */
   gnutls_x509_trust_list_t trust_list;
+  GGnutlsCertificateCredentials *credentials;
 
   /*
    * These are hash tables of GBytes -> GPtrArray<GBytes>. The values of
@@ -76,6 +80,41 @@ G_DEFINE_TYPE_WITH_CODE (GTlsDatabaseGnutls, g_tls_database_gnutls, G_TYPE_TLS_D
                                                 g_tls_database_gnutls_initable_interface_init);
                          );
 
+GGnutlsCertificateCredentials *
+g_gnutls_certificate_credentials_new (GError **error)
+{
+  GGnutlsCertificateCredentials *credentials = g_new (GGnutlsCertificateCredentials, 1);
+  int ret;
+
+  ret = gnutls_certificate_allocate_credentials (&credentials->credentials);
+  if (ret != 0)
+    {
+      g_set_error (error, G_TLS_ERROR, G_TLS_ERROR_MISC, _("Failed to allocate credentials: %s"), gnutls_strerror (ret));
+      g_free (credentials);
+      return NULL;
+    }
+
+  g_atomic_ref_count_init (&credentials->ref_count);
+  return credentials;
+}
+
+GGnutlsCertificateCredentials *
+g_gnutls_certificate_credentials_ref (GGnutlsCertificateCredentials *credentials)
+{
+  g_atomic_ref_count_inc (&credentials->ref_count);
+  return credentials;
+}
+
+void
+g_gnutls_certificate_credentials_unref (GGnutlsCertificateCredentials *credentials)
+{
+  if (g_atomic_ref_count_dec (&credentials->ref_count))
+    {
+      g_clear_pointer (&credentials->credentials, gnutls_certificate_free_credentials);
+      g_free (credentials);
+    }
+}
+
 static GHashTable *
 bytes_multi_table_new (void)
 {
@@ -226,6 +265,7 @@ g_tls_database_gnutls_finalize (GObject *object)
   g_clear_pointer (&priv->handles, g_hash_table_destroy);
 
   gnutls_x509_trust_list_deinit (priv->trust_list, 1);
+  g_clear_pointer (&priv->credentials, g_gnutls_certificate_credentials_unref);
 
   g_mutex_clear (&priv->mutex);
 
@@ -485,8 +525,6 @@ g_tls_database_gnutls_verify_chain (GTlsDatabase             *database,
   GTlsCertificateFlags result;
   guint gnutls_result;
   CertificateChain *gnutls_chain;
-  const char *hostname = NULL;
-  char *free_hostname = NULL;
   int gerr;
 
   g_return_val_if_fail (G_IS_TLS_CERTIFICATE_GNUTLS (chain),
@@ -511,23 +549,10 @@ g_tls_database_gnutls_verify_chain (GTlsDatabase             *database,
 
   result = g_tls_certificate_gnutls_convert_flags (gnutls_result);
 
-  if (G_IS_NETWORK_ADDRESS (identity))
-    hostname = g_network_address_get_hostname (G_NETWORK_ADDRESS (identity));
-  else if (G_IS_NETWORK_SERVICE (identity))
-    hostname = g_network_service_get_domain (G_NETWORK_SERVICE (identity));
-  else if (G_IS_INET_SOCKET_ADDRESS (identity))
-    {
-      GInetAddress *addr;
-
-      addr = g_inet_socket_address_get_address (G_INET_SOCKET_ADDRESS (identity));
-      hostname = free_hostname = g_inet_address_to_string (addr);
-    }
-  if (hostname)
-    {
-      if (!gnutls_x509_crt_check_hostname (gnutls_chain->chain[0], hostname))
-        result |= G_TLS_CERTIFICATE_BAD_IDENTITY;
-      g_free (free_hostname);
-    }
+  if (identity)
+    result |= g_tls_certificate_gnutls_verify_identity (G_TLS_CERTIFICATE_GNUTLS (chain),
+                                                        identity,
+                                                        error);
 
   certificate_chain_free (gnutls_chain);
   return result;
@@ -577,6 +602,74 @@ g_tls_database_gnutls_populate_trust_list (GTlsDatabaseGnutls        *self,
   return gerr >= 0;
 }
 
+static gnutls_x509_trust_list_t
+create_trust_list (GTlsDatabaseGnutls  *self,
+                   GError             **error)
+{
+  GTlsDatabaseGnutlsClass *database_class = G_TLS_DATABASE_GNUTLS_GET_CLASS (self);
+  gnutls_x509_trust_list_t trust_list;
+  int ret;
+
+  ret = gnutls_x509_trust_list_init (&trust_list, 0);
+  if (ret != 0)
+    {
+      g_set_error (error, G_TLS_ERROR, G_TLS_ERROR_MISC, "Failed to initialize trust list: %s", gnutls_strerror (ret));
+      return NULL;
+    }
+
+  g_assert (database_class->populate_trust_list);
+  if (!database_class->populate_trust_list (self, trust_list, error))
+    {
+      gnutls_x509_trust_list_deinit (trust_list, TRUE);
+      return NULL;
+    }
+
+  return trust_list;
+}
+
+GGnutlsCertificateCredentials *
+g_tls_database_gnutls_get_credentials (GTlsDatabaseGnutls  *self,
+                                       GError             **error)
+{
+  GTlsDatabaseGnutlsPrivate *priv = g_tls_database_gnutls_get_instance_private (self);
+  GGnutlsCertificateCredentials *credentials;
+  gnutls_x509_trust_list_t trust_list = NULL;
+  GError *my_error = NULL;
+
+  g_mutex_lock (&priv->mutex);
+  if (priv->credentials)
+    goto out;
+
+  credentials = g_gnutls_certificate_credentials_new (&my_error);
+  if (!credentials)
+    {
+      g_propagate_error (error, my_error);
+      goto out;
+    }
+
+  /* The gnutls_certificate_credentials_t takes ownership of the
+   * gnutls_x509_trust_list_t, so we have to wastefully create a new one here
+   * rather than reuse priv->trust_list.
+   *
+   * Alternatively, we could steal priv->trust_list, unsetting it in the priv
+   * struct, to avoid duplicating this work. But this would violate the rule
+   * that our private data is read-only after initialization. Let's not.
+   */
+  trust_list = create_trust_list (self, error);
+  if (!trust_list)
+    {
+      g_gnutls_certificate_credentials_unref (credentials);
+      goto out;
+    }
+
+  gnutls_certificate_set_trust_list (credentials->credentials, trust_list, 0);
+
+  priv->credentials = g_steal_pointer (&credentials);
+out:
+  g_mutex_unlock (&priv->mutex);
+  return priv->credentials;
+}
+
 static void
 g_tls_database_gnutls_class_init (GTlsDatabaseGnutlsClass *klass)
 {
@@ -611,14 +704,9 @@ g_tls_database_gnutls_initable_init (GInitable     *initable,
   if (g_cancellable_set_error_if_cancelled (cancellable, error))
     return FALSE;
 
-  gnutls_x509_trust_list_init (&trust_list, 0);
-
-  g_assert (G_TLS_DATABASE_GNUTLS_GET_CLASS (self)->populate_trust_list);
-  if (!G_TLS_DATABASE_GNUTLS_GET_CLASS (self)->populate_trust_list (self, trust_list, error))
-    {
-      result = FALSE;
-      goto out;
-    }
+  trust_list = create_trust_list (self, error);
+  if (!trust_list)
+    return FALSE;
 
   subjects = bytes_multi_table_new ();
   issuers = bytes_multi_table_new ();
@@ -658,7 +746,6 @@ g_tls_database_gnutls_initable_init (GInitable     *initable,
       g_mutex_unlock (&priv->mutex);
     }
 
-out:
   if (trust_list)
     gnutls_x509_trust_list_deinit (trust_list, 1);
   if (subjects)
